using FitConnect.Encryption;
using FitConnect.Services.Interfaces;
using FitConnect.Services.Models.v1.Case;
using FitConnect.Services.Models.v1.Submission;
using Microsoft.Extensions.Logging;
using Microsoft.IdentityModel.Tokens;
using Newtonsoft.Json;
using NJsonSchema;

namespace FitConnect.Services;

internal class SubmissionService : RestCallService, ISubmissionService {
    private readonly string _baseUrl;
    private readonly ILogger? _logger;
    private readonly IOAuthService _oAuthService;
    private readonly JsonWebKey? _signatureValidationKey;

    public SubmissionService(string baseUrl, IOAuthService oAuthService,
        JsonWebKey? signatureValidationKey,
        string version = "v1", ILogger? logger = null) : base(
        $"{baseUrl}/{version}", logger) {
        _baseUrl = baseUrl;
        _oAuthService = oAuthService;
        _signatureValidationKey = signatureValidationKey;
        _logger = logger;
    }

    /// <summary>
    ///     <para>@PostMapping("/submissions")</para>
    /// </summary>
    /// <param name="submissionDto">RequestBody</param>
    /// <returns></returns>
    public async Task<SubmissionCreatedDto> CreateSubmission(CreateSubmissionDto submissionDto) {
        await _oAuthService.EnsureAuthenticatedAsync();
        var result = await RestCall<SubmissionCreatedDto>(
            "/submissions",
            HttpMethod.Post,
            JsonConvert.SerializeObject(submissionDto));

        if (result == null)
            throw new ArgumentException("Submission creation failed");

        return result;
    }


    /// <summary>
    ///     <para>
    ///         @PutMapping(value = "/submissions/{submissionId}/attachments/{attachmentId}", consumes =
    ///         "application/jose")
    ///     </para>
    /// </summary>
    /// <param name="submissionId">PathVariable</param>
    /// <param name="attachmentId">PathVariable</param>
    /// <param name="encryptedAttachmentContent">RequestBody</param>
    /// <returns></returns>
    public async Task<bool> AddSubmissionAttachment(string submissionId, string attachmentId,
        string encryptedAttachmentContent) {
        await _oAuthService.EnsureAuthenticatedAsync();
        await RestCallForString(
            $"/submissions/{submissionId}/attachments/{attachmentId}",
            HttpMethod.Put,
            encryptedAttachmentContent,
            "application/jose");
        return true;
    }

    /// <summary>
    ///     <para>@PutMapping(value = "/submissions/{submissionId}", consumes = "application/json") </para>
    /// </summary>
    /// <param name="submissionId">PathVariable</param>
    /// <param name="submitSubmission">RequestBody</param>
    /// <returns></returns>
    public async Task<SubmissionReducedDto?> SubmitSubmission(string submissionId,
        SubmitSubmissionDto submitSubmission) {
        await _oAuthService.EnsureAuthenticatedAsync();
        var body = JsonConvert.SerializeObject(submitSubmission, new JsonSerializerSettings {
            NullValueHandling = NullValueHandling.Ignore,
            Formatting = Formatting.Indented
        });

        var result = await RestCall<SubmissionReducedDto>($"/submissions/{submissionId}",
            HttpMethod.Put,
            body);

        return result;
    }


    /// <summary>
    ///     <para>@GetMapping("/submissions")</para>
    /// </summary>
    /// <param name="destinationId">RequestParam</param>
    /// <param name="offset">RequestParam</param>
    /// <param name="limit">RequestParam</param>
    /// <returns></returns>
    public async Task<SubmissionsForPickupDto> ListSubmissionsAsync(string? destinationId,
        int offset,
        int limit) {
        await _oAuthService.EnsureAuthenticatedAsync();
        var url = destinationId != null
            ? $"/submissions?destinationId={destinationId}&offset={offset}&limit={limit}"
            : $"/submissions?offset={offset}&limit={limit}";
        return (await RestCall<SubmissionsForPickupDto>(url, HttpMethod.Get)) ??
               new SubmissionsForPickupDto();
    }

    /// <summary>
    ///     <para>@GetMapping("/submissions/{submissionId}")</para>
    /// </summary>
    /// <param name="submissionId">PathVariable</param>
    /// <returns></returns>
    public async Task<SubmissionDto> GetSubmissionAsync(string submissionId) {
        await _oAuthService.EnsureAuthenticatedAsync();
        var result =
            (await RestCall<SubmissionDto>($"/submissions/{submissionId}", HttpMethod.Get)) ??
            new SubmissionDto();

        return result;
    }


    /// <summary>
    ///     <para>
    ///         @GetMapping(value = "/submissions/{submissionId}/attachments/{attachmentId}", produces =
    ///         "application/jose")
    ///     </para>
    /// </summary>
    /// <param name="submissionId">PathVariable</param>
    /// <param name="attachmentId">PathVariable</param>
    /// <returns></returns>
    public async Task<string> GetAttachmentAsync(string submissionId, string attachmentId) {
        await _oAuthService.EnsureAuthenticatedAsync();
        var result = await RestCallForString($"/submissions/{submissionId}/attachments/{attachmentId}",
            HttpMethod.Get,
            null,
            "application/jose",
            "application/jose");

        return result;
    }

    public string GetKey(string keyId) {
        throw new NotImplementedException();
    }

    public async Task<List<string>?> GetStatusForSubmissionAsync(string caseId,
        string destinationId,
        bool skipTest = false) {
        await _oAuthService.EnsureAuthenticatedAsync();
        var events = await RestCall<EventLogDto>($"/cases/{caseId}/events", HttpMethod.Get);
        if (events == null)
            return null;

        var valid = await ValidateSignature(events, destinationId);
        valid &= await ValidateSchema(events);

        if (!valid) {
            _logger?.LogError("Invalid SET, signature can not be verified");
            if (events.EventLog != null)
                _logger?.LogTrace("Tested events: {Events}",
                    events.EventLog.Aggregate((a, b) => a + "\n" + b));
            if (!skipTest)
                throw new InvalidOperationException("Invalid SET, signature can not be verified");
        }

        return events.EventLog?.ToList();
    }


    private async Task<bool> ValidateSchema(EventLogDto events, bool mandatory = false) {
        var valid = true;
        if (events.EventLog == null)
            return true;

        foreach (var eventString in events.EventLog) {
            var eventContent = Base64UrlEncoder.Decode(eventString.Split('.')[1]);
            _logger?.LogTrace("Validating event: {Event}", eventContent);
            var payload = JsonConvert.DeserializeObject<Dictionary<string, object>>(eventContent);
            if (payload == null) {
                _logger?.LogWarning("Invalid event: {Event}", eventContent);
                continue;
            }

            if (payload.ContainsKey("$schema")) {
                var schema = (string)((dynamic)payload)["$schema"];
                _logger?.LogInformation("Validating event against schema: {Schema}", schema);

                try {
                    var schemaRaw = await RestCallForString(new Uri(schema), HttpMethod.Get);
                    var schemaOk = JsonHelper.ValidateJsonSchema(schemaRaw, eventContent);


                    if (mandatory)
                        valid = false;


                    if (!schemaOk) {
                        _logger?.Log(mandatory ? LogLevel.Error : LogLevel.Warning,
                            "Error on validating JSON Schema");

                        if (mandatory)
                            throw new ArgumentException("Invalid JSON Schema for SET event");
                    }
                    else {
                        _logger?.LogDebug("SET event is valid");
                    }
                }
                catch (Exception e) {
                    _logger?.LogError(e, "Error validating event against schema {Exception}", e);
                    if (mandatory)
                        valid = false;
                }
            }
        }

        return valid;
    }

    private async Task<bool> ValidateSignature(EventLogDto events, string destinationId) {
        var keys = (await GetJsonWebKeysForEvent(events, destinationId)).ToList();

        if (events.EventLog == null)
            return true;

        var result = true;
        foreach (var s in events.EventLog) {
            var header = JsonConvert.DeserializeObject<Dictionary<string, object>>(
                Base64UrlEncoder.Decode(s.Split('.')[0]));

            if (header == null)
                throw new ArgumentException("Invalid event header");

            var kid = (string)header["kid"];
            result &= FitEncryption.VerifyJwt(s, keys.Where(k => k.Kid == kid), _logger);
        }

        var valid = result;

        if (!valid) {
            _logger?.LogDebug("Signature is invalid");
            _logger?.LogDebug("Tested with keys: {Keys}",
                keys.Aggregate("", (a, b) => $"{b.Kid}, {a}"));
        }

        return valid;
    }

    private async Task<IEnumerable<JsonWebKey>> GetJsonWebKeysForEvent(EventLogDto events,
        string destinationId) {
        var keySet = new JsonWebKeySet(await Router.GetSubmissionServiceValidationJwk(_baseUrl));
        var keys = (_signatureValidationKey == null
            ? keySet.Keys
            : keySet.Keys.Append(_signatureValidationKey)).ToList();


        var result =
            (await GetKeyIdsFromEvent(events, destinationId, keys.Select(k => k.Kid).ToList()))
            .Union(keys).ToList();

        var valid = result.Aggregate(true,
            (a, key) => (new CertificateHelper(_logger).ValidateCertificate(key)));
        if (!valid) {
            _logger?.LogError("(SET Events verification) Invalid certificate");
            throw new ArgumentException("(SET Events verification) Invalid certificate");
        }
        else {
            _logger?.LogInformation("(SET Events verification) Certificate is valid");
        }

        return result;
    }

    private async Task<IEnumerable<JsonWebKey>> GetKeyIdsFromEvent(EventLogDto events,
        string destinationId, ICollection<string> knownKeys) {
        if (events.EventLog == null)
            return new List<JsonWebKey>();

        // Load Key from GET {{submission_api_url}}/v1/destinations/{{destinationId}}/keys/{{keyId}}
        var keyIds = events.EventLog.Select(ExtractSubmissionIdFromEvent).ToList();

        var result = new List<JsonWebKey>();
        foreach (var (submission, keyId) in keyIds) {
            if (knownKeys.Contains(keyId))
                continue;

            var keyJson = await RestCallForString($"/destinations/{destinationId}/keys/{keyId}",
                HttpMethod.Get);
            result.Add(new JsonWebKey(keyJson));
        }

        return result;
    }

    private (string submissionId, string keyId) ExtractSubmissionIdFromEvent(string events) {
        var jwtParts = events.Split('.').Select(Base64UrlEncoder.Decode).ToList();

        var header = JsonConvert.DeserializeObject<Dictionary<string, object>>(jwtParts[0]);
        var payload = JsonConvert.DeserializeObject<Dictionary<string, object>>(jwtParts[1]);

        if (header == null || payload == null)
            throw new NullReferenceException("Event string is invalid");

        var keyId = (string)header["kid"];
        var submissionId = ((string)payload["sub"]).Split(':')[1];

        return (submissionId, keyId);
    }
}
